# Functions for the simulation study in the paper "A Unifying Approach to the 
# Estimation of the Conditional Akaike Information in Generalized Linear Mixed Models"
# The simulation study is started with starter.r
# Exponential distribution with random intercept.
# author: B. Saefken

tr <- function(X){
	return(sum(diag(X)))
}

genData <- function(NoCluster, NoSize, sigma2, beta0, beta1, NoSets){
  # Generates the data
  #
  # Args: NoCluster = Number of clusters
  #		    NoSize    = Size of the clusters
  #       sigma2    = Variance of random effects
  #       beta0     = Intercept (fixed)
  #       beta1     = Slope (fixed)
  #       NoSets    = The number of data sets
  #
  # Returns:	res = A list of data.frames
  #
	res <-list(NoCluster  = NoCluster, 
             NoSize     = NoSize, 
             NoSets     = NoSets, 
             sigma2     = sigma2, 
             beta0      = beta0, 
             beta1      = beta1)	
	for(j in 1:NoSets){	
		randomEffects <- rep(rnorm(NoCluster, 0, sigma2), each = NoSize)
		Z <- exp(randomEffects + rep(beta0 + beta1 * ((1:NoSize)), NoCluster))
		Y <- numeric(length(Z))
		for(i in 1:length(Z)){	
			Y[i] <- rgamma(1, scale = Z[i], shape = 1) 
		}	
		daten <- data.frame(Y, 
                        rep(1:NoSize, NoCluster), 
                        as.factor(rep(1:NoCluster, each = NoSize)), 
                        randomEffects)	
		names(daten) <- c("y", "x", "groups", "randomEffects")	
		res[[paste("dataSet", j, sep = "")]] <- daten
	}
	return(res)		
}

generateDataSets <- function(Slops, Intercepts, anzahlCluster, clusterGroesse, varianzRandomEffects, anzahlDaten){
  # A wrapper that calls genData for different settings
  #
  # Args: Slops                = Vector of slops
  #		    Intercepts           = Vector of intercepts
  #       anzahlCluster        = Vector of number of clusters
  #       clusterGroesse       = Vector of cluster sizes
  #       varianzRandomEffects = Vector of variances of the random effects
  #       anzahlDaten          = Vector of number of data sets
  #
  # Returns:	res = A list of data.frames
  #
  for(w in Slops){
    for(v in Intercepts){
      for(p in anzahlCluster){
        for(r in clusterGroesse){
          for(s in varianzRandomEffects){		
          liste <- genData(NoCluster = p, 
                          NoSize = r, 
                          sigma2 = s, 
                          beta0 = v, 
                          beta1 = w, 
                          NoSets = anzahlDaten)
          speicherString <- paste("data/NC_", p, 
                                  "_CG_", r, 
                                  "_t2_", s, 
                                  "_b0_", v, 
                                  "_b1_", w, 
                                  ".Rdata", sep = "")
          save(liste, file = speicherString)
          }
        }
      }
    }
  }
  }

margAIC <- function(daten){	
  # Calculates the marginal AIC for a data.frame generated by genData with the 
  # help of Laplace approximation
  #
  # Args: daten = data.frame generated by genData
  #
  # Returns:	mAIC = The marginal AIC corresponding to a GLMM with daten
  #
	daten.mgcv <- try(gamm(y ~ x + s(groups, bs = "re"), 
                    family = Gamma(link = "log"), 
                    data = daten, 
                    scale = 1, 
                    verbosePQL = FALSE)) 
	if (class(daten.mgcv) == "try-error") {
		cat("NA's produziert")
		return(NA)
	}
	mu    <- daten.mgcv$gam$fitted.values
	tau2  <- daten.mgcv$gam$smooth[[1]]$S.scale/daten.mgcv$gam$sp[[1]]
  
	Z <- predict(daten.mgcv$gam,type="lpmatrix")[, -c(1, 2)]
  W <- diag(rep(1, length(mu)))
	b <- daten.mgcv$gam$coefficients[-c(1, 2)]
  
	loglik <- try(sum(dgamma(daten$y, 
                           shape = 1, 
                           scale = mu, 
                           log = TRUE)) - 
                ncol(Z) * log(tau2)/2 - crossprod(b)/(2*tau2) - 
                log(det(t(Z) %*% W %*% Z + 1/tau2 * diag(nrow = ncol(Z))))/2)
	if (class(loglik) == "try-error") {
		cat("NA's produziert")
		loglik <- NA
		return(NA)
	}
	mAIC <- -2 * loglik + 2 * 3
	return(mAIC)
}

simpAIC <- function(daten){
  # Calculates the simple AIC for a data.frame generated by genData with bias
  # correction equal to the fixed effects (2)
  #
  # Args: daten = data.frame generated by genData
  #
  # Returns:	sAIC = The AIC corresponding to a GLM with daten
  #
	b <- try(glm(y ~ x, family = Gamma(link = "log"), data = daten))
	if (class(b)[1] == "try-error") {
		cat("NA's produziert in simpAIC \n")
		return(NA)
	}
	ll <- sum(dgamma(daten$y, shape = 1, scale = b$fitted.values, log = TRUE))
	return(-2 * ll + 2 * 2)
}

yuyauAIC <- function(daten){
  # Calculates the conditional AIC of Yu & Yau (2012) for a data.frame generated
  # by genData. Adapted from the MATLAB code of the authors.
  #
  # Args: daten = data.frame generated by genData
  #
  # Returns:	yAIC = The conditional AIC of Yu & Yau (2012) corresponding to a 
  #                  GLMM with daten
  #
	daten.mgcv <- try(gamm(y ~ x + s(groups, bs = "re"), 
                         family = Gamma(link = "log"), 
                         data = daten, 
                         scale = 1, 
                         verbosePQL = FALSE))
	if (class(daten.mgcv)[1] == "try-error") {
		cat("NA's produziert in yyAIC 1\n")
		return(NA)
	}	
	thetaHat <- -1/daten.mgcv$gam$fitted.values
	mu       <- daten.mgcv$gam$fitted.values
	tau2     <- daten.mgcv$gam$smooth[[1]]$S.scale/daten.mgcv$gam$sp[[1]]
  
	Z <- predict(daten.mgcv$gam, type = "lpmatrix")[,-c(1, 2)]
	B <- diag(rep(1, length(mu)))
	b <- daten.mgcv$gam$coefficients[-c(1, 2)]
	X <- model.matrix(daten.mgcv$gam)[, 1:2]
  
	H11 <- t(X) %*% B %*% X
	H12 <- t(X) %*% B %*% Z
	H21 <- t(Z) %*% B %*% X
	H22 <- t(Z) %*% B %*% Z + 1/tau2 * diag(nrow = ncol(Z))
  
	Hthetatheta <- rbind(cbind(H11, H12), cbind(H21, H22))
	Htautheta   <- as.matrix(t(c(0, 0, b)) / (tau2^2))
	Hthetatau   <- t(Htautheta)
	H22_inv     <- try(solve(H22))
  
	if (class(H22_inv) == "try-error") {
		cat("NA's produziert in yyAIC 2\n")
		return(NA)
	}
	Htautau <- 1/2 * tr(-1 / (tau2^4) * H22_inv %*% H22_inv + 2 / (tau2^3) * 
             H22_inv) - ncol(Z) / (2 * tau2^2) + t(b) %*% b / (tau2^3)
	Hstern <- rbind(cbind(H11, H12), cbind(H21, t(Z) %*% B %*% Z))
	Htautau_inv <- try(solve(Htautau))
	if (class(Htautau_inv)=="try-error") {
		cat("NA's produziert in yyAIC 3\n")
		return(NA)
	}
	H <- try(solve(Hthetatheta - Hthetatau %*% Htautau_inv %*% Htautheta))
	if (class(H) == "try-error") {
		cat("NA's produziert in yyAIC 4\n")
		return(NA)
	}
	rhohat <- tr(H %*% Hstern)
	loglik <- sum(-daten$y / mu - log(mu))		
	yAIC   <- -2 * loglik + 2 * rhohat
	return(yAIC)
}

condAIC <- function(daten){
  # Calculates the conditional AIC proposed in the paper for a data.frame 
  # generated by genData. The bias correction is obtained by numerical 
  # integration
  #
  # Args: daten = data.frame generated by genData
  #
  # Returns:	cAIC = The conditional AIC corresponding to a GLMM with daten
  #
	daten.mgcv <- try(gamm(y ~ x + s(groups, bs = "re"), 
                         family = Gamma(link = "log"), 
                         data = daten, scale = 1, 
                         verbosePQL = FALSE))
	if (class(daten.mgcv)[1] == "try-error") {
		cat("NA's produziert in condAIC 1\n")
		return(NA)
	}
	thetaHat <- -1/daten.mgcv$gam$fitted.values	
	mu <- daten.mgcv$gam$fitted.values
	k <-0		
	for(i in 1:length(daten$y)){
		estimatortheta <- function(z){
			daten2 <- daten				
			daten2$y[i] <-z			
			daten2.mgcv <- try(gamm(y ~ x + s(groups, bs = "re"), 
                              family = Gamma(link = "log"), 
                              data = daten2, scale = 1, 
                              verbosePQL = FALSE))
			if (class(daten2.mgcv)[1] == "try-error"){
				cat("NA's produziert in condAIC 2\n")
				return(NA)
			}
			esttheta <- -1/predict(daten2.mgcv$gam, daten2, type = "response")[[i]]
		return(esttheta)	
	}	
		vestimatortheta <- Vectorize(estimatortheta)
		value <- try(integrate(vestimatortheta, lower = 0, upper = daten$y[i])$value)
		if (class(value)[1] == "try-error") {
			cat("NA's produziert in condAIC 3\n")
			k <- NA
			return(NA)
			} else {
			k <- k + value
			}
	}
	BC <- sum(daten$y * thetaHat) - k
	condloglik <- sum(-log(mu) - daten$y/mu)
	cAIC <- -2 * condloglik + 2 * BC
return(cAIC)
}

DonAIC <- function(daten){
  # Calculates the conditional AIC proposed by Donohue et al. (2010) for a 
  # data.frame generated by genData.
  #
  # Args: daten = data.frame generated by genData
  #
  # Returns:	DAIC = The conditional AIC corresponding to a GLMM with daten
  #
	b <- try(gamm(y ~ x + s(groups, bs = "re"), 
                family = Gamma(link = "log"),
                data = daten, 
                scale = 1, 
                verbosePQL = FALSE))
	if (class(b)[1] == "try-error") {
		cat("NA's produziert in DonAIC\n")
		return(NA)
	}	
	DonBC  <- sum(b$gam$edf)
	loglik <- sum(dgamma(daten$y, 
                       shape = 1, 
                       scale = b$gam$fitted.values,
                       log = TRUE))
	DAIC <- - 2 * loglik + 2 * DonBC
	return(DAIC)
}

nullAIC <- function(daten){
  # Calculates the conditional AIC proposed in the paper for a data.frame 
  # generated by genData for a GLM. The bias correction is obtained by numerical 
  # integration
  #
  # Args: daten = data.frame generated by genData
  #
  # Returns:	cAIC = The conditional AIC corresponding to a GLM with daten
  #
	b <- try(glm(y ~ x, family = Gamma(link = "log"), data = daten))
	if(class(b) == "try-error"){
		cat("NA's produziert")
		return(NA)
	}	
	thetaHat <- -1/b$fitted.values
	k <-0		
	for(i in 1:length(daten$y)){
		estimatortheta <- function(z){
			daten2 <- daten				
			daten2$y[i] <- z			
			b2 <- try(glm(y ~ x, family = Gamma(link = "log"), data = daten2))
			if (class(b2)[1] == "try-error") {
				cat("NA's produziert in condAIC 2\n")
				return(NA)
			}
			esttheta <- -1/predict(b2, daten2, type="response")[[i]]
		return(esttheta)	
	}	
		vestimatortheta <- Vectorize(estimatortheta)
		value <- try(integrate(vestimatortheta, lower = 0, upper = daten$y[i])$value)
		if (class(value)[1] == "try-error") {
			cat("NA's produziert in condAIC 3\n")
			k <- NA
			return(NA)
			} else {
			k <- k + value
			}
	}
	BC <- sum(daten$y * thetaHat) - k
	ll <- sum(dgamma(daten$y, 
                   shape = 1, 
                   scale = b$fitted.values, 
                   log   = TRUE))
	AIC <- - 2 * ll + 2 * BC
	return(AIC)
}

Simulation <- function(Slops, Intercepts, anzahlCluster, clusterGroesse, varianzRandomEffects){
  # The simulation study. 
  #
  # Args: Slops = A vector of slops.
  #       Intercepts = a vector of intercepts.
  #       anzahlCluster = A vector of the number of clusters.
  #       clusterGroesse = A vector of cluster sizes.
  #       varianzRandomEffects = A vector of variances of the random effects.
  #
  # Returns:	Writes a table into the results folder.
  #
for(w in Slops){
	for(v in Intercepts){
		for(p in anzahlCluster){
			for(r in clusterGroesse){
				for(s in varianzRandomEffects){					
				cat("Start of loop: Slope: ", w, 
            "; Intercept: ", v, 
            "; Anzahl Cluster: ", p, 
            "; Cluster Groesse: ", r, 
            "; tau^2: ", s, 
            "\n", sep = "")
				loadString <- paste("data/NC_", p, 
                            "_CG_", r, 
                            "_t2_", s, 
                            "_b0_", v, 
                            "_b1_", w,
                            ".Rdata", sep = "")
				load(file = loadString)
				
				m0AIC <- foreach(k = 1:liste$NoSets, .combine = "rbind")%dopar%{
					nullAIC(liste[[6 + k]])
        }
				
        yyAIC <- foreach(k = 1:liste$NoSets, .combine = "rbind")%dopar%{
        	yuyauAIC(liste[[6 + k]])
        }
				
				cAIC <- foreach(k = 1:liste$NoSets, .combine = "rbind")%dopar%{
					condAIC(liste[[6 + k]])
				}
				
				mAIC <- foreach(k = 1:liste$NoSets, .combine = "rbind")%dopar%{
					margAIC(liste[[6 + k]])
        }
				
				dAIC <- foreach(k = 1:liste$NoSets, .combine = "rbind")%dopar%{
				 	DonAIC(liste[[6 + k]])
        }
        
				sAIC <- foreach(k = 1:liste$NoSets, .combine = "rbind")%dopar%{
				 	simpAIC(liste[[6 + k]])
        }        
				
				if (file.exists("results") == FALSE) {
          dir.create("results")	
				}
				tableString <- paste("results/AIC_Tabelle_NoCluster_", p, 
                             "_NoSize_", r, 
                             "_sigma2_", s, 
                             "_Int_", v, 
                             "_Slope_", w, 
                             ".txt", sep = "")			
				write.table(data.frame(m0AIC, m0AIC, cAIC, yyAIC, dAIC, mAIC), 
                    file      = tableString, 
                    quote     = FALSE, 
                    row.names = FALSE)	
				}
			}
		}
	}
}
}
